import ipywidgets as widgets
from IPython.display import clear_output, display
from IPython.display import Image
from .env import Environment
from .pr import Project
from pprint import pprint
from io import _io
import plantuml
import pandas as pd

class SystemDescription:
  def __init__(this, env: Environment):
    this.env = env

    this.compTypes = ['Category', 'Component', 'ControlAction', 'DomainSet',
                      'Exit', 'Feedback', 'Function', 'Link', 'UseCase']

    this.pr = Project(env)
    this.pr.FetchSchema()
    this.pr.FetchEntities(this.compTypes)
    this.pr.FetchStructure()

    # last function processed - when parsing call structure
    this.lastFunction = ""

    this.output = {}

    # item: Control Action & Feedback types
    this.itemList = []
    this.itemByName = {}

    this.arcList = []
    this.arcByName = {}

    this.output = {}
    this.csImage = {}

    # control actions
    this.caDF = {}
    this.caDT = {}

    # use cases
    this.ucDF = {}
    this.ucDT = {}
    return

  # create control structure plantuml diagram file
  def ControlStructureDiagram(this):
    db = this.pr.entities
    dbDict = this.pr.entitiesDict

    lineColor = '#tan;line.bold;text:black'

    with open('./diagrams/cs.txt', 'w') as f:
      f.write('skinparam roundCorner 15\n')
      f.write('scale .75\n')
      f.write('@startuml\n')

      contextId = dbDict['System Context']

      # every control structure starts with 'System Context' component
      for bf_component in db[contextId]['rels']['built from']:
        this._output_built_from(f, 0, bf_component)

      this._create_control_structure_arcs()
      for arc in this.arcList:
        arrowDir = ""
        labelDecorator = ""
        if this.arcByName[arc]['type'] == "ControlAction":
          arrowDir = ' -do-> '
          labelDecorator = '<&caret-bottom>'
        else:
          arrowDir = ' -up-> '
          labelDecorator = '<&caret-top>'

        # concatenate arc labels
        labelCount = 0
        labelString = ""
        for label in this.arcByName[arc]['arcLabels']:
          labelCount += 1
          labelString += labelDecorator + label
          if labelCount < len(this.arcByName[arc]['arcLabels']):
            labelString += '\\n'

        labelString += '\n'
        f.write(this.arcByName[arc]['outputComponent'] +
                arrowDir +
                this.arcByName[arc]['inputComponent'] +
                ' ' + lineColor + ' : ' + labelString)

      f.write('@enduml\n')

    server = plantuml.PlantUML('http://www.plantuml.com/plantuml/img/')
    try:
      ret = server.processes_file('./diagrams/cs.txt',
                            './diagrams/cs.png', './diagrams/cs_error.html')
    except BaseException as err:
      print(err)

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)
    this.csImage = Image('./diagrams/cs.png')
    # display cs diagram to output area
    with this.output:
      display(this.csImage)
    return

# recursively output built from
  def _output_built_from(this, f: _io.TextIOWrapper, level: int, component: dict):
    db = this.pr.entities
    dbDict = this.pr.entitiesDict
    csCompCatId = dbDict['SD: CS: Components']
    compColor = '#deepskyblue'
    sysColor = '#white;line:grey;line.bold'

    # check if component is to be included in control structure (by category)
    found = False
    for comp in db[csCompCatId]['rels']['categorizes']:
      if comp['targetId'] == component['targetId']:
        found = True
        break
    if found != True:
      return
    compType = db[component['targetId']]['attrs']['type']['value']
    compName = db[component['targetId']]['attrs']['title']['value']
    compAbbr = db[component['targetId']]['attrs']['abbreviation']['value']

    if compType == 'Human':
      f.write(' '*level + 'person "' + compName + '" as ' +
              compAbbr + ' ' + compColor )
    elif compType == "System":
      f.write(' '*level + 'rectangle "' + compName + '" as ' +
              compAbbr + ' ' + sysColor )
    else:
      f.write(' '*level + 'rectangle "' + compName + '" as ' +
              compAbbr + ' ' + compColor )

    # check for built from components
    if 'built from' in db[component['targetId']]['rels']:
      # check that at least one built from component is
      # included in control structure
      for bf_comp in db[component['targetId']]['rels']['built from']:
        found = False
        for comp in db[csCompCatId]['rels']['categorizes']:
          if comp['targetId'] == bf_comp['targetId']:
            found = True
            break
        if found == True:
          break
      # at least on 'built from' component in control structure
      if found == True:
        f.write(' {\n')
        level += 2
        for bf_component in db[component['targetId']]['rels']['built from']:
          this._output_built_from(f, level, bf_component)
        level -= 2
        f.write(' '*level + '}\n')
      else:
        f.write('\n')
    else:
      f.write('\n')
    return

  def _create_control_structure_arcs(this):
    this.arcByName = {}
    this.arcList = []
    this.itemByName = {}
    this.itemList = []
    db = this.pr.entities
    dbDict = this.pr.entitiesDict
    csFuncCatId = dbDict['SD: CS: Functions']

    # fetch output control actions and feedback items
    for function in db[csFuncCatId]['rels']['categorizes']:
      if 'outputs' in db[function['targetId']]['rels']:
        for output in db[function['targetId']]['rels']['outputs']:
          itemEntry = {}
          itemEntry['ItemTitle'] = db[output['targetId']]['attrs']['title']['value']
          itemEntry['itemDescription'] = db[output['targetId']]['attrs']\
                    ['description']['value']
          itemEntry['itemType'] = db[output['targetId']]['type']
          itemEntry['inputComponent'] = ""
          outCompId = db[function['targetId']]['rels']['allocated to'][0]
          itemEntry['outputComponent'] = db[outCompId['targetId']]['attrs']\
                    ['abbreviation']['value']

          # add to list and dict
          this.itemList.append(itemEntry['ItemTitle'])
          this.itemByName[itemEntry['ItemTitle']] = itemEntry

    # fetch input control actions and feedback items
    for function in db[csFuncCatId]['rels']['categorizes']:
      if 'inputs' in db[function['targetId']]['rels']:
        for input in db[function['targetId']]['rels']['inputs']:
          inCompId = db[function['targetId']]['rels']['allocated to'][0]
          itemTitle = db[input['targetId']]['attrs']['title']['value']
          # update input component
          this.itemByName[itemTitle]['inputComponent'] = \
                  db[inCompId['targetId']]['attrs']['abbreviation']['value']

    this.itemList.sort()

    # combine items with same output/input pair into an arc
    for item in this.itemList:
      arcName = this.itemByName[item]['outputComponent'] + ':' + \
                  this.itemByName[item]['inputComponent']

      if arcName in this.arcByName.keys():
        this.arcByName[arcName]['arcLabels'].append(item)
      else:
        this.arcList.append(arcName)
        arcEntry = {}
        arcEntry['type'] = this.itemByName[item]['itemType']
        arcEntry['outputComponent'] = this.itemByName[item]['outputComponent']
        arcEntry['inputComponent'] = this.itemByName[item]['inputComponent']
        arcLabels = []
        arcLabels.append(item)
        arcEntry['arcLabels'] = arcLabels
        this.arcByName[arcName] = arcEntry

    return

  # display use case diagram
  def UseCaseDiagrams(this):
    db = this.pr.entities
    dbDict = this.pr.entitiesDict

    # get category ID for System Description: Use Cases
    ucCatId = dbDict['SD: UC']

    # retrieve categorized Use Cases
    for uc in db[ucCatId]['rels']['categorizes']:
      ucName = db[uc['targetId']]['attrs']['name']['value']
      inputPath = "./diagrams/" + ucName.replace(" ", "") + ".txt"
      outputPath = "./diagrams/" + ucName.replace(" ", "") + ".png"
      errorPath = "./diagrams/" + ucName.replace(" ", "") + "_error.html"

      with open(inputPath, 'w') as f:
        f.write('@startuml\n')
        f.write('left to right direction\n')

        # retrieve 'involves' Components (actors)
        for comp in db[uc['targetId']]['rels']['involves']:
          f.write('actor "' + db[comp['targetId']]['attrs']['title']['value'] + '" as ' +
                   db[comp['targetId']]['attrs']['abbreviation']['value'] + '\n')

        # retrieve 'describes' Component (single association [0])
        desId = db[uc['targetId']]['rels']['describes'][0]['targetId']
        f.write('package ' + db[desId]['attrs']['title']['value'] + ' {\n')

        # retrieve 'includes' UseCases
        for iuc in db[uc['targetId']]['rels']['includes']:
          f.write('  usecase "' + db[iuc['targetId']]['attrs']['name']['value'] + '" as ' +
                  db[iuc['targetId']]['attrs']['number']['value'] + '\n')

        f.write('}\n')

        # retrieve 'involves' components per use case (connections)
        for iuc in db[uc['targetId']]['rels']['includes']:
          for invuc in db[iuc['targetId']]['rels']['involves']:

            # create connection between actors and usecase with display 'hint' u/d/l/r
            f.write(db[invuc['targetId']]['attrs']['abbreviation']['value'] + ' -' +
                    invuc['attrs']['hint']['value'] + '-> ' +
                    db[iuc['targetId']]['attrs']['number']['value']   + '\n')

        f.write('@enduml\n')

      # create UML diagram
      server = plantuml.PlantUML('http://www.plantuml.com/plantuml/img/')
      try:
        ret = server.processes_file(inputPath, outputPath, errorPath)
      except BaseException as err:
        print(err)

      # setup output area
      this.output = widgets.Output(layout={'border': '1px solid black'})
      display(this.output)
      # display uc diagram to output area
      with this.output:
        display(Image(outputPath))

  # display use case table
  def UseCaseTable(this):
    db = this.pr.entities
    dbDict = this.pr.entitiesDict
    # get category ID for System Description: Use Cases
    ucCatId = dbDict['SD: UC']

    ucTable = []
    # retrieve categorized Use Cases
    for uc in db[ucCatId]['rels']['categorizes']:
      for uci in db[uc['targetId']]['rels']['includes']:
        ucItem = []
        ucItem.append(db[uci['targetId']]['attrs']['name']['value'])
        ucItem.append(db[uci['targetId']]['attrs']['description']['value'])
        ucItem.append(db[uci['targetId']]['attrs']['preconditions']['value'])
        ucItem.append(db[uci['targetId']]['attrs']['postconditions']['value'])
        ucTable.append(ucItem)

    this.ucDF = pd.DataFrame(ucTable, columns = ['Name', 'Description',
                  'Pre-Conditions', 'Post-Conditions'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.ucDT = data_table.DataTable(this.ucDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.ucDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.ucDF)
    return

  # display use case message sequence diagrams
  def MSCDiagrams(this):
    db = this.pr.entities
    dbDict = this.pr.entitiesDict
    structures = this.pr.structures

    # get category ID for System Description: Use Cases
    ucCatId = dbDict['SD: UC']

    # retrieve categorized Use Cases
    for uc in db[ucCatId]['rels']['categorizes']:
      # retrieve 'includes' UseCases
      for iuc in db[uc['targetId']]['rels']['includes']:
        # each usecase is 'categorized by' a single category that
        # includes an 'ordered' list of participants (swimlanes)
        catById = db[iuc['targetId']]['rels']['categorized by'][0]

        # need to sort the parts by 'order' relationship attribute
        partDict = {}
        for part in db[catById['targetId']]['rels']['categorizes']:
          if db[part['targetId']]['type'] == 'Component':
            partDict[part['attrs']['order']['value']] = part['targetId']

        sortedPartList = sorted(partDict)
        ucName = db[iuc['targetId']]['attrs']['name']['value']
        inputPath = "./diagrams/" + ucName.replace(" ", "") + ".txt"
        outputPath = "./diagrams/" + ucName.replace(" ", "") + ".png"
        errorPath = "./diagrams/" + ucName.replace(" ", "") + "_error.html"

        with open(inputPath, 'w') as f:
          f.write('@startuml\n')
          f.write('title MSC: ' + ucName + '\n')

          # output participants (swimlanes)
          for part in sortedPartList:
            if db[partDict[part]]['attrs']['type']['value'] == 'Human':
              f.write('actor "' + db[partDict[part]]['attrs']['title']['value'] +
                      '" as ' + db[partDict[part]]['attrs']['abbreviation']['value'] + '\n')
            else:
              f.write('control "' + db[partDict[part]]['attrs']['title']['value'] +
                      '" as ' + db[partDict[part]]['attrs']['abbreviation']['value'] + '\n')
          # parse call structure for included use case 'elaborated by' function
          constructs = structures[db[iuc['targetId']]['rels']['elaborated by'][0]\
                        ['targetId']]['mainBranch']['constructs']
          this._process_constructs(f, 0, constructs)
          f.write('@enduml\n')

        # create UML diagram
        server = plantuml.PlantUML('http://www.plantuml.com/plantuml/img/')
        try:
          ret = server.processes_file(inputPath, outputPath, errorPath)
        except BaseException as err:
          print(err)

        # setup output area
        this.output = widgets.Output(layout={'border': '1px solid black'})
        display(this.output)
        # display uc diagram to output area
        with this.output:
          display(Image(outputPath))
      return

  def _process_constructs(this, f: _io.TextIOWrapper, level: int, constructs: dict):
    db = this.pr.entities

    for construct in constructs:
      if construct['constructType'] == 'function':
        this.lastFunction = construct['functionId']
        this._process_function(f, level, construct['functionId'])
        # process function exits
        exitCount = 0
        for branch in construct['branches']:
          exitCount += 1
          exitName = db[branch['exitId']]['attrs']['name']['value']
          if exitCount == 1:
            f.write(' '*level + 'alt ' + exitName + '\n')
          else:
            f.write(' '*level + 'else ' + exitName + '\n')
          level += 2
          this._process_constructs(f, level, branch['constructs'])
          level -= 2
        if exitCount > 0:
          f.write(' '*level + 'end\n')
      elif construct['constructType'] == 'parallel':
        f.write(' '*level + 'par\n')
        for branch in construct['branches']:
          level += 2
          this._process_constructs(f, level, branch['constructs'])
          level -= 2
        f.write(' '*level + 'end par\n')
      elif construct['constructType'] == 'loop':
        if construct['annotation'] is not None:
          f.write(' '*level + 'loop ' + construct['annotation'] + '\n')
        else:
          f.write(' '*level + 'loop\n')

        level += 2
        this._process_constructs(f, level, construct['branch']['constructs'])
        level -= 2
        f.write(' '*level + 'end loop\n')
      elif construct['constructType'] == 'replicate':
        # retrieve domain set name
        dsName = db[construct['domainSetId']]['attrs']['name']['value']
        f.write(' '*level + 'group repl[' + dsName + ']\n')

        level += 2
        this._process_constructs(f, level, construct['branch']['constructs'])
        level -= 2
        f.write(' '*level + 'end\n')
      elif construct['constructType'] == 'loopExit':
        comp = db[this.lastFunction]['rels']['allocated to'][0]['targetId']
        compAbbrv = db[comp]['attrs']['abbreviation']['value']
        f.write(' '*level + compAbbrv + ' --> ' + compAbbrv + ' : [Loop Exit]\n')
      else:
        print("Unknown Construct Type!")
    return

  def _process_function(this, f: _io.TextIOWrapper, level: int, functionId: str):
    db = this.pr.entities

    # process 'outputs' items (control action / feedback)
    if 'outputs' in db[functionId]['rels']:
      for output in db[functionId]['rels']['outputs']:
        itemType = db[output['targetId']]['type']
        itemName = db[output['targetId']]['attrs']['name']['value']
        # retrieve 'input to' endpoint
        inFunc = db[output['targetId']]['rels']['input to'][0]['targetId']
        # retrieve function 'allocated to' component abbreviation
        outComp = db[functionId]['rels']['allocated to'][0]['targetId']
        outAbbrv = db[outComp]['attrs']['abbreviation']['value']
        inComp = db[inFunc]['rels']['allocated to'][0]['targetId']
        inAbbrv = db[inComp]['attrs']['abbreviation']['value']

        if itemType == 'ControlAction':
          f.write(' '*level + outAbbrv + ' -> ' + inAbbrv +
                  ' : <&caret-bottom>' + itemName + '\n')
        else:
          f.write(' '*level + outAbbrv + ' -> ' + inAbbrv +
                  ' : <&caret-top>' + itemName + '\n')
    return

  # display table of control actions
  def ControlActionTable(this):

    caTable = []
    for ca in this.itemList:
      caItem = []
      if this.itemByName[ca]['itemType'] == 'ControlAction':
        caItem.append(ca)
        caItem.append(this.itemByName[ca]['itemDescription'])
        caTable.append(caItem)

    this.caDF = pd.DataFrame(caTable, columns = ['Control Action', 'Description'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.caDT = data_table.DataTable(this.caDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.caDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.caDF)
    return



from .env import Environment, Section
import requests
from pprint import pprint
import pandas as pd
import ipywidgets as widgets
from ipywidgets import Layout, Button, VBox, FloatText, Textarea
from ipywidgets import Dropdown, Label, IntSlider
from IPython.display import clear_output, display
import json


class Exercises:
  def __init__(this, env: Environment, section: Section):
    this.env = env
    this.section = section
    this.category = ""
    this.exList = []
    this.exDict = {}

    this.exDF = {}
    this.exDT = {}

    # form widget handles
    this.output = {}
    this.title = {}
    this.status = {}
    this.decision = {}
    this.save = {}
    this.result = {}
    this.table = {}
    this.vbox = {}

    if this.section.name == Section.RISK_ASSESSMENT.name:
      this.category = "EX: Risk Assessment"
    else:
      print("Invalid Exercises Section")
      return

    status, catId = this.env.getCategoryId(this.category)
    if status != 200:
      print("Failed to fetch assessment category: " + str(status))
      return

    # fetch 'categorized' exercises
    r = requests.get(this.env.url + 'projects/' +
                this.env.projectDict[this.env.project.value] +
                '/entities/' + catId +
                '/relationshiptargets/' + this.env.relationDict['categorizes'] +
                '?sortBlockId=' + this.env.sortBlockDict['Numeric'],
                allow_redirects=False, headers=this.env.header)
    if r.status_code != 200:
      print("Failed to fetch categorized Exercises: " + str(r.status_code))
      return

    # collect all exercises
    exercises = r.json()['results']
    for exercise in exercises:
      exTitle = ""
      exEntry = {}
      exEntry['entityDefinitionId'] = exercise['entityDefinitionId']
      exEntry['folderId'] = exercise['folderId']
      exEntry['entityId'] = exercise['id']
      exEntry['exDecision'] = " "
      for attr in exercise['attributes']:
        if attr['definitionId'] == this.env.attributeDict['name']:
          if attr['value'] is not None:
            exEntry['exTitle'] = attr['value']['value']
        if attr['definitionId'] == this.env.attributeDict['description']:
          if attr['value'] is not None:
            exEntry['exDescription'] = attr['value']['value']['plainText']
        if attr['definitionId'] == this.env.attributeDict['note:status']:
          if attr['value'] is not None:
            exEntry['exStatus'] = attr['value']['value']['value']
        if attr['definitionId'] == this.env.attributeDict['note:decision']:
          if attr['value'] is not None:
            exEntry['exDecision'] = attr['value']['value']['plainText']
            if len(exEntry['exDecision']) == 0:
              exEntry['exDecision'] = " "

      this.exList.append(exEntry['exTitle'])
      this.exDict[exEntry['exTitle']] = exEntry

    this.exList.sort()

  def Edit(this):

    this.title = Dropdown(
      options=this.exList,
      description='Title:',
      disabled=False
    )
    this.status = Dropdown(
      options=['Open', 'Closed'],
      description='Status:',
      disabled=False
    )
    this.decision = Textarea(
      description='Decision:',
      placeholder='Enter Decision...',
      disabled=False
    )
    this.save = Button(
      description='Save',
      tooltip='Click to Save status & decision.',
      button_style='success',
      disabled=False
    )
    this.save.on_click(this._on_save)
    this.result = widgets.Text(
      value='',
      placeholder='Save Result',
      description='',
      disabled=True)

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)
    this._refresh_output()

  def _on_save(this, b):
    # update exercise response
    payload = {}
    attrList = [{},{}]
    payload['attributes'] = attrList
    payload['entityDefinitionId'] = this.exDict[this.title.value]['entityDefinitionId']
    payload['folderId'] = this.exDict[this.title.value]['folderId']

    attrList[0]['definitionId'] = this.env.attributeDict['note:status']
    attrList[0]['value'] = {}
    attrList[0]['value']['value'] = this.status.value
    attrList[0]['value']['dataType'] = "string"
    attrList[1]['definitionId'] = this.env.attributeDict['note:decision']
    attrList[1]['value'] = {}
    attrList[1]['value']['value'] = {}
    attrList[1]['value']['value']['plainText'] = this.decision.value
    attrList[1]['value']['dataType'] = "text"

    this.env._set_json_header()
    r = requests.put(this.env.url + 'projects/' +
                    this.env.projectDict[this.env.project.value] +
                    '/entities/' + this.exDict[this.title.value]['entityId'],
                    allow_redirects=False, headers=this.env.header,
                    json=payload)
    if r.status_code != 200:
      this.result.value("Failed to save Exercise update: " + str(r.status_code))
      return
    else:
      this.result.value = "Save Success!"
      # update local dictionary
      this.exDict[this.title.value]['exStatus'] = this.status.value
      this.exDict[this.title.value]['exDecision'] = this.decision.value

      this._refresh_output()

  def _refresh_output(this):

    # assemble exercise table
    exTable = []
    for ex in this.exList:
      exItem = []
      exItem.append(this.exDict[ex]['exTitle'])
      exItem.append(this.exDict[ex]['exDescription'])
      exItem.append(this.exDict[ex]['exStatus'])
      exItem.append(this.exDict[ex]['exDecision'])
      exTable.append(exItem)

    this.exDF = pd.DataFrame(exTable, columns = ['Title', 'Description', \
                                                  'Status', 'Decision'])
    with this.output:
      clear_output()

    # output the DF/DT plus edit form
    try:
      from google.colab import data_table
      data_table.enable_dataframe_formatter()
      this.exDT = data_table.DataTable(this.exDF, include_index=False)
      # Display dataframa via Colab datatable
      this.table = widgets.Output(layout={'border': '1px solid black'})
      with this.table:
        display(this.exDT)
    except ModuleNotFoundError:
      # Display basic dataframe
      this.table = widgets.Output(layout={'border': '1px solid black'})
      with this.table:
        display(this.exDF)
    finally:
      vbox_items = [this.table, this.title, this.status, \
                    this.decision, this.save, this.result]
      this.vbox = VBox(vbox_items)
      with this.output:
        display(this.vbox)


import ipywidgets as widgets
import logging

class MANBLogHandler(logging.Handler):
  """ Custom logging handler sending logs to an output widget """

  def __init__(self, *args, **kwargs):
    super(MANBLogHandler, self).__init__(*args, **kwargs)
    layout = {
      'width': '100%',
      'height': '100%',
      'border': '1px solid black'
    }
    self.out = widgets.Output(layout=layout)

  def emit(self, record):
    """ Overload of logging.Handler method """
    formatted_record = self.format(record)
    new_output = {
      'name': 'stdout',
      'output_type': 'stream',
      'text': formatted_record+'\n'
    }
    self.out.outputs = (new_output, ) + self.out.outputs

  def show_logs(self):
    """ Show the logs """
    display(self.out)

  def clear_logs(self):
    """ Clear the current logs """
    self.out.clear_output()




import requests
from .env import Environment
from pprint import pprint
from IPython.display import clear_output

# Project db - schema definition & entities
class Project:
  def __init__(this, env: Environment):
    this.env = env

    this.sortBlockDict = {}
    this.sortBlockDict['Numeric'] = '78b85fc0-3aec-442d-95b9-2b291bd4d1bc'

    # schema definition dictionaries name <-> uuid

    # fetch: entityType UUID
    #   n2IdEntityDict['<entityTypeName>']['id']
    # fetch: entity AttributeType UUID
    #   n2IdEntityDict['<entityTypeName>']['attrs']['<attributeTypeName>']['id']
    this.n2idEntityDict = {}

    # fetch: entityType Name
    #   id2nEntityDict['<entityTypeUUID']['name']
    # fetch: entity AttributeType Name
    #   id2nEntityDict['<entityTypeUUID']['attrs']['<attributeTypeUUID>']['name']
    this.id2nEntityDict = {}

    # same as above for relationships
    this.id2nRelDict = {}
    this.n2idRelDict = {}

    # entity data store

    # fetch: entity type
    #   entities[<entity uuid>]['type']
    # fetch: entity attribute value
    #   entities[<entity uuid>]['attrs'][<attr name>]
    # fetch: entity related targets list
    #   entities[<entity uuid>]['rels'][<rel name>]
    this.entities = {}

    # entity name -> uuid
    this.entitiesDict = {}

    # entity type list
    # fetch: list of entities by type
    #   entitesForTypeList[<entityTypeName>]
    this.entitiesForTypeList = {}

    # call structure for functions
    this.structures = {}

    return

  # fetch and save entities
  def FetchEntities(this, entityList: list):
    clear_output()
    print("Fetch Entities....")
    this.entities.clear()

    for entity in entityList:
      entityEntry = {}
      # fetch entity instances
      this.env._set_json_header()
      r = requests.get(this.env.url + 'projects/' +
                  this.env.projectDict[this.env.project.value] + '/entities/byclass/' +
                  this.n2idEntityDict[entity]['id'] +
                  '?sortBlockId=' + this.sortBlockDict['Numeric'],
                  allow_redirects=False, headers=this.env.header)
      if r.status_code != 200:
        print("Failed to fetch entity: " + entity + ": " + str(r.status_code))
        return

      entityInstances = r.json()['results']

      # initialize entity list for entity Type
      this.entitiesForTypeList[entity] = []

      for entityInstance in entityInstances:
        entityEntry = {}
        entityId = entityInstance['id']
        entityEntry['type'] = this.id2nEntityDict[entityInstance['entityDefinitionId']]['name']

        # add entity to list by type
        this.entitiesForTypeList[entity].append(entityId)

        entityEntry['attrs'] = {}
        entityEntry['rels'] = {}
        for attr in entityInstance['attributes']:
          attrEntry = this._parse_attr_value(attr)

          # set attribute name
          entityEntry['attrs'][this.id2nEntityDict[entityInstance\
            ['entityDefinitionId']]['attrs'][attr['definitionId']]] = attrEntry


          if attr['definitionId'] == this.n2idEntityDict[entity]['attrs']['name']:
            # create entity name to uuid dictionary
            this.entitiesDict[attrEntry['value']] = entityId

        this.entities[entityId] = entityEntry

    # fetch relationship instances
    this.env._set_json_header()
    r = requests.get(this.env.url + 'projects/' +
                this.env.projectDict[this.env.project.value] +
                '/relationships',
                allow_redirects=False, headers=this.env.header)
    if r.status_code != 200:
      print("Failed to fetch relationships: " + str(r.status_code))
      return

    relationships = r.json()['results']
    for relationship in relationships:
      # check if both ends of relationship exist in project scope
      if (relationship['sourceEntityId'] in this.entities) and \
            (relationship['targetEntityId'] in this.entities):
        relName = this.id2nRelDict[relationship['relationDefinitionId']]['name']
        relRevName = this.id2nRelDict[relationship['relationDefinitionId']]['rev-name']

        attrs = {}

        # add relationship attributes
        if len(relationship['attributes']) > 0:
          for attr in relationship['attributes']:
            attrEntry = this._parse_attr_value(attr)

            # set attribute name
            attrs[this.id2nRelDict[relationship\
              ['relationDefinitionId']]['attrs'][attr['attributeDefinitionId']]] = attrEntry

        # add relationsip to source entity
        relEntry = {}
        relEntry['targetId'] = ""
        relEntry['attrs'] = attrs
        relEntry['targetId'] = relationship['targetEntityId']
        if relName not in this.entities[relationship['sourceEntityId']]['rels']:
          this.entities[relationship['sourceEntityId']]['rels'][relName] = []

        this.entities[relationship['sourceEntityId']]['rels'][relName].append(relEntry)

        # add reverse relationship to target entity
        relRevEntry = {}
        relRevEntry['targetId'] = ""
        relRevEntry['attrs'] = attrs
        relRevEntry['targetId'] = relationship['sourceEntityId']
        if relRevName not in this.entities[relationship['targetEntityId']]['rels']:
          this.entities[relationship['targetEntityId']]['rels'][relRevName] = []

        this.entities[relationship['targetEntityId']]['rels'][relRevName].append(relRevEntry)

    clear_output()
    print("Entites Fetch Complete!")
    return

  # fetch and save schema mapping - for name to/from uuid mapping
  def FetchSchema(this):
    clear_output()
    print("Fetch Schema Definition....")

    # fetch project schema
    this.env._set_json_header()
    r = requests.get(this.env.url + 'projects/' +
                this.env.projectDict[this.env.project.value] + '/schemas/',
                allow_redirects=False, headers=this.env.header)
    if r.status_code != 200:
      print("Failed to fetch schema: " + str(r.status_code))
      return
    schema = r.json()

    # build entity defition dictionaries
    for entityDefinition in schema['entityDefinitions']:
      entityNEntry = {}
      entityNEntry['id'] = entityDefinition['id']
      entityNEntry['attrs'] = {}
      this.n2idEntityDict[entityDefinition['name']] = entityNEntry

      entityIEntry = {}
      entityIEntry['name'] = entityDefinition['name']
      entityIEntry['attrs'] = {}
      this.id2nEntityDict[entityDefinition['id']] = entityIEntry

      for attrDefinition in entityDefinition['attributeDefinitions']:
        entityNEntry['attrs'][attrDefinition['name']] = attrDefinition['id']
        entityIEntry['attrs'][attrDefinition['id']] = attrDefinition['name']

    # build relations definition dictionary
    for relationshipDefinition in schema['relationshipDefinitions']:

      relationshipNEntry = {}
      relationshipNEntry['id'] = relationshipDefinition['relationDefinitionA']['id']
      relationshipNEntry['rev-id'] = relationshipDefinition['relationDefinitionB']['id']
      relationshipNEntry['attrs'] = {}
      this.n2idRelDict[relationshipDefinition['relationDefinitionA']['name']] = relationshipNEntry
      # add relationship attributes
      for attrDefinition in relationshipDefinition['attributeDefinitions']:
        relationshipNEntry['attrs'][attrDefinition['name']] = attrDefinition['id']

      relationshipNEntry = {}
      relationshipNEntry['id'] = relationshipDefinition['relationDefinitionB']['id']
      relationshipNEntry['rev-id'] = relationshipDefinition['relationDefinitionA']['id']
      relationshipNEntry['attrs'] = {}
      this.n2idRelDict[relationshipDefinition['relationDefinitionB']['name']] = relationshipNEntry
      # add relationship attributes
      for attrDefinition in relationshipDefinition['attributeDefinitions']:
        relationshipNEntry['attrs'][attrDefinition['name']] = attrDefinition['id']

      relationshipIEntry = {}
      relationshipIEntry['name'] = relationshipDefinition['relationDefinitionA']['name']
      relationshipIEntry['rev-name'] = relationshipDefinition['relationDefinitionB']['name']
      relationshipIEntry['attrs'] = {}
      this.id2nRelDict[relationshipDefinition['relationDefinitionA']['id']] = relationshipIEntry
      # add relationship attributes
      for attrDefinition in relationshipDefinition['attributeDefinitions']:
        relationshipIEntry['attrs'][attrDefinition['id']] = attrDefinition['name']

      relationshipIEntry = {}
      relationshipIEntry['name'] = relationshipDefinition['relationDefinitionB']['name']
      relationshipIEntry['rev-name'] = relationshipDefinition['relationDefinitionA']['name']
      relationshipIEntry['attrs'] = {}
      this.id2nRelDict[relationshipDefinition['relationDefinitionB']['id']] = relationshipIEntry
      # add relationship attributes
      for attrDefinition in relationshipDefinition['attributeDefinitions']:
        relationshipIEntry['attrs'][attrDefinition['id']] = attrDefinition['name']

    clear_output()
    print("Schema Definition Fetch Complete!")
    return

  # parse entity / relationship attribute value
  def _parse_attr_value(this, attr: dict) -> dict:

    attrEntry = {}

    # set attribute name
    if attr['value'] is not None:
      attrEntry['dataType'] = attr['value']['dataType']
    else:
      attrEntry['dataType'] = ""

    # retrieve attribute value
    if attrEntry['dataType'] == "text":
      attrEntry['value'] = attr['value']['value']['plainText']
    elif attrEntry['dataType'] == "string":
      attrEntry['value'] = attr['value']['value']
    elif attrEntry['dataType'] == "hierarchicalNumber":
      attrEntry['value'] = attr['value']['value']
    elif attrEntry['dataType'] == "integer":
      attrEntry['value'] = attr['value']['value']
    elif attrEntry['dataType'] == "enumeration":
      attrEntry['value'] = attr['value']['value']['value']
    else:
      attrEntry['value'] = ""

    return attrEntry

  def FetchStructure(this):
    clear_output()
    print("Fetch Structure Definition....")
    this.structures.clear()

    # retrieve structure for functions
    for entity in this.entitiesForTypeList['Function']:
      # fetch structure
      this.env._set_json_header()
      r = requests.get(this.env.url + 'projects/' +
                  this.env.projectDict[this.env.project.value] + '/structures/' +
                  entity, allow_redirects=False, headers=this.env.header)
      if r.status_code != 200:
        print("Failed to fetch structure: " + entity + ": " + str(r.status_code))
        return

      this.structures[entity] = r.json()

    clear_output()
    print("Structure Definition Fetch Complete!")



from .env import Environment
from .sd import SystemDescription
from .pr import Project
import requests
from IPython.display import clear_output, display
import sys
from pprint import pprint
import pandas as pd
import ipywidgets as widgets


class RiskAssessment:
  def __init__(this, env: Environment):
    this.env = env

    this.compTypes = ['Category', 'ControlAction', 'Hazard', 'Loss', 'HazardousAction']

    this.pr = Project(env)
    this.pr.FetchSchema()
    this.pr.FetchEntities(this.compTypes)

    this.output = {}

    # losses
    this.lossDF = {}
    this.lossDT = {}

    # hazards
    this.hazardDF = {}
    this.hazardDT = {}

    # hazardous control action
    this.hcaDF = {}
    this.hcaDT = {}

    # control action analysis
    this.caaDF = {}
    this.caaDT = {}

    this.hcaTypes = ['Providing', 'NotProviding', 'TooEarlyTooLate']
    return

  def LossTable(this):
    db = this.pr.entities
    dbTypeList = this.pr.entitiesForTypeList


    lossTable = []
    for loss in dbTypeList['Loss']:
      lossItem = []
      lossItem.append(db[loss]['attrs']['number']['value'])
      lossItem.append(db[loss]['attrs']['name']['value'])
      lossItem.append(db[loss]['attrs']['priority']['value'])

      hazardList = " "
      if 'ma: is caused by' in db[loss]['rels']:
        first = True
        for hazard in db[loss]['rels']['ma: is caused by']:
          hazardNum = db[hazard['targetId']]['attrs']['number']['value']
          if first == True:
            first = False
          else:
            hazardList += ','
          hazardList += hazardNum
      lossItem.append(hazardList)
      lossTable.append(lossItem)

    this.lossDF = pd.DataFrame(lossTable, columns = ['ID', 'Title', 'Priority', \
                                        'is caused by: Hazard'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.lossDT = data_table.DataTable(this.lossDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.lossDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.lossDF)

  def HazardTable(this):
    db = this.pr.entities
    dbTypeList = this.pr.entitiesForTypeList

    hazardTable = []
    for hazard in dbTypeList['Hazard']:
      hazardItem = []
      hazardItem.append(db[hazard]['attrs']['number']['value'])
      hazardItem.append(db[hazard]['attrs']['name']['value'])
      hazardItem.append(db[hazard]['attrs']['description']['value'])

      lossList = " "
      if 'ma: leads to' in db[hazard]['rels']:
        first = True
        for loss in db[hazard]['rels']['ma: leads to']:
          lossNum = db[loss['targetId']]['attrs']['number']['value']
          if first == True:
            first = False
          else:
            lossList += ','
          lossList += lossNum
      hazardItem.append(lossList)

      hcaList = " "
      if 'ma: is caused by' in db[hazard]['rels']:
        first = True
        for hca in db[hazard]['rels']['ma: is caused by']:
          hcaNum = db[hca['targetId']]['attrs']['number']['value']
          if first == True:
            first = False
          else:
            hcaList += ','
          hcaList += hcaNum
      hazardItem.append(hcaList)

      hazardTable.append(hazardItem)

    this.hazardDF = pd.DataFrame(hazardTable, columns = ['ID', 'Title', 'Description', \
                    'leads to: Loss', 'is caused by: Hazardous Action'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.hazardDT = data_table.DataTable(this.hazardDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.hazardDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.hazardDF)

  def HazardousActionTable(this):
    db = this.pr.entities
    dbTypeList = this.pr.entitiesForTypeList

    hcaTable = []
    for hca in dbTypeList['HazardousAction']:
      hcaItem = []
      hcaItem.append(db[hca]['attrs']['number']['value'])
      hcaItem.append(db[hca]['attrs']['name']['value'])
      hcaItem.append(db[hca]['attrs']['description']['value'])
      hcaItem.append(db[hca]['attrs']['variationType']['value'])

      hazardList = " "
      if 'ma: leads to' in db[hca]['rels']:
        first = True
        for hazard in db[hca]['rels']['ma: leads to']:
          hazardNum = db[hazard['targetId']]['attrs']['number']['value']
          if first == True:
            first = False
          else:
            hazardList += ','
          hazardList += hazardNum
      hcaItem.append(hazardList)

      caVariationList = " "
      if 'ma: variation of' in db[hca]['rels']:
        first = True
        for caVariation in db[hca]['rels']['ma: variation of']:
          caVariationNum = db[caVariation['targetId']]['attrs']['number']['value']
          if first == True:
            first = False
          else:
            caVariationList += ','
          caVariationList += caVariationNum
      hcaItem.append(caVariationList)

      hcaTable.append(hcaItem)

    this.hcaDF = pd.DataFrame(hcaTable, columns = ['ID', 'Title', 'Description', \
                    'Variation Type', 'leads to: Hazard', 'variation of: Control Action'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.hcaDT = data_table.DataTable(this.hcaDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.hcaDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.hcaDF)

  def ControlActionAnalysisTable(this):
    db = this.pr.entities
    dbTypeList = this.pr.entitiesForTypeList

    caaTable = []
    for ca in dbTypeList['ControlAction']:
      for hcaType in this.hcaTypes:
        caItem = []
        caItem.append(db[ca]['attrs']['title']['value'])
        caItem.append(hcaType)
        # check if hca of this type exists for this ca
        if 'ma: has variation' in db[ca]['rels']:
          for hca in db[ca]['rels']['ma: has variation']:
            if db[hca['targetId']]['attrs']['variationType']['value'] == hcaType:
              caItem.append(db[hca['targetId']]['attrs']['number']['value'])
            else:
              caItem.append(' ')
        else:
          caItem.append(' ')

        # Todo: save / retrieve variation type justification
        caItem.append(' ')
        caaTable.append(caItem)

    this.caaDF = pd.DataFrame(caaTable, columns = ['Control Action', 'Variation', \
                    'has variation: Hazardous Action', 'has variation: .justification'])

    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)

    with this.output:
      try:
        from google.colab import data_table
        data_table.enable_dataframe_formatter()
        this.caaDT = data_table.DataTable(this.caaDF, include_index=False)
        # Display dataframa via Colab datatable
        display(this.caaDT)
      except ModuleNotFoundError:
        # Display basic dataframe
        display(this.caaDF)




from .env import Environment
from .ra import RiskAssessment
import requests
from IPython.display import clear_output, display
import sys
from pprint import pprint
import pandas as pd
import ipywidgets as widgets

class VulnerabilityAssessment:
  def __init__(this, env: Environment, ra: RiskAssessment):
    this.env = env
    this.ra = ra
    clear_output()
    print("Initializing Vulnerability Assessment")

  def LossScenarioTable(this):
    print("Loss Table")





import requests
import ipywidgets as widgets
from ipywidgets import VBox
import subprocess
from IPython.display import clear_output, display
from typing import Tuple
import urllib.request
import os
from enum import Enum
from .log import MANBLogHandler
import logging
import json
from pprint import pprint

class Section(Enum):
  RISK_ASSESSMENT = 1
  VULNERABILITY_ASSESSMENT = 2

class Environment:

  def __init__(this, domain: str, path: str):

    clear_output()
    # REST API base URL
    this.domain = domain
    this.path = path
    this.url = domain + path

    # GENESYS API refresh token
    this.refresh_token = ""

    # API content types
    this.content_form = 'application/x-www-form-urlencoded'
    this.content_json = 'application/json'
    # REST API Header
    #  cf-access-token: -> tunnel token
    #  Authorization: -> GENESYS api token
    this.header = {'cf-access-token': "",
                  'Authorization': "",
                  'Accept': 'application/json',
                  'Content-Type': ""}
    # Login widget handles
    this.output = {}

    this.name = {}
    this.pwd = {}
    this.login = {}
    this.status = {}
    this.project = {}
    this.vbox = {}
    this.vboxitems = []

    # Authorized Projects
    this.projectList = []
    this.projectDict = {}

    # Setup MANB Logger
    this.logger = logging.getLogger(__name__)
    this.loghandler = MANBLogHandler()
    this.loghandler.setFormatter(logging.Formatter(\
                '%(asctime)s - [%(levelname)s] %(message)s'))
    this.logger.addHandler(this.loghandler)
    this.logger.setLevel(logging.ERROR)

    # Create temporary directory for plantuml output
    if not os.path.exists('./diagrams'):
      os.makedirs('./diagrams')

    # Create temporary directory for local project storage
    if not os.path.exists('./projects'):
      os.makedirs('./projects')

    # form to select local / remote repository
    this.repository = widgets.ToggleButtons(
      options=['Local ', 'Remote '],
      value='Local ',
      description='Repository:',
      disabled=False,
      #button_style='info',
      tooltips=['Read-Only via Github JSON File',
                'Read-Write via GENESYS REST-API'],
      icons=['file','database']
    )
    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)
    # display project selection widgets to output area
    with this.output:
      display(this.repository)

    return

  # Authenticate the Cloudflare Tunnel
  def Tunnel(this):

    clear_output()
    if this.repository.value == 'Local ':
      print("Tunnel authentication not required for local repository.")
    else:
      # Download cloudflared CLI
      print("Downloading Tunnel Client...")
      this._get_cloudflared()
      clear_output()

      process = subprocess.Popen(['./cloudflared', 'access', 'login', this.domain],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True)

      while True:
        # stderr output is URL for authentication (if token not already cached)
        output = process.stderr.readline()
        print(output.strip())
        # Wait for cloudflared to complete
        return_code = process.poll()
        if return_code is not None:
          if return_code == 0:
            # Line 1: Successfully fetched your token:
            # Line 2: <token>
            count = 0
            for output in process.stdout.readlines():
              if output.strip() != "":
                count += 1
                if count == 2:
                  this.header['cf-access-token'] = output.strip()
                  clear_output()
                  print("Tunnel Token Fetch Success!")
          else:
            print("Cloudflared Error Exit: " + str(return_code))
          break

  # Login to GENESYS
  def Login(this):

    this.project = widgets.Select(
      options=[''],
      value='',
      description='Project:',
      disabled=False)

    if this.repository.value == 'Local ':
      # no login for local repository
      this.vbox_items = [this.project]

      # retrieve local project list
      if not os.path.exists('./projects/projects.json'):
        url = 'https://raw.githubusercontent.com/tsherburne/de-textbook/' \
              'main/projects/projects.json'
        urllib.request.urlretrieve(url, './projects/projects.json')

      with open('./projects/projects.json') as f:
        projects = json.load(f)

      # build project list and dict
      this.projectList = []
      this.projDict = {}
      for proj in projects:
        this.projectList.append(proj['name'])
        this.projectDict[proj['name']] = proj['path']

      this.projectList.sort()

      this.project.options = this.projectList
      this.project.value = this.projectList[0]

    else:
      this.name = widgets.Text(
        value='',
        placeholder='Enter username',
        description='User Name:',
        disabled=False)

      this.pwd = widgets.Password(
        value='',
        placeholder='Enter password',
        description='Password:',
        disabled=False)

      this.login = widgets.Button(
        description='Login',
        tooltip='Enter User Name and Password - then click to Login.',
        icon='sign-in'
      )

      this.status = widgets.Text(
        value='',
        placeholder='Login Status',
        description='',
        disabled=True)
      # Register Login button callback
      this.login.on_click(this._on_login)
      this.vbox_items = [this.name, this.pwd, this.login, this.status, this.project]

    this.vbox = VBox(this.vbox_items)
    # setup output area
    this.output = widgets.Output(layout={'border': '1px solid black'})
    display(this.output)
    # display login widgets to output area
    with this.output:
      display(this.vbox)

  # Login button clicked
  def _on_login(this, b):

    # Authenticate name / password
    payload = {'grant_type': 'password',
               'username': this.name.value,
               'password': this.pwd.value}

    this._set_form_header()
    r = requests.post(this.url + 'token', data=payload,
                      allow_redirects=False, headers=this.header)
    # clear password
    del payload
    this.pwd.value = ""

    # clear project list
    this.projectList = []
    this.projectDict = {}
    this.project.options = ['']
    this.project.value = ''

    if r.status_code != 200:
      this.header['Authorization'] = ""
      this.status.value = 'Login Failed: ' + str(r.status_code)
      return

    this.status.value = 'Login Success!'
    # save access token for future requests
    this.header['Authorization'] = 'bearer ' + r.json()['access_token']

    # get authorized projects for user
    r = requests.get(this.url + 'projects', allow_redirects=False,
                      headers=this.header)
    if r.status_code != 200:
      this.status.value = 'Project Request Failed: ' + str(r.status_code)
      return

    projects = r.json()['results']

    # build project list and dict
    for proj in projects:
      this.projectList.append(proj['name'])
      this.projectDict[proj['name']] = proj['id']

    this.projectList.sort()

    this.project.options = this.projectList
    this.project.value = this.projectList[0]

  # Set REST header for json content
  def _set_json_header(this):
    this.header['Content-Type'] = this.content_json
  # Set REST header for form content
  def _set_form_header(this):
    this.header['Content-Type'] = this.content_form

  # Download Cloudflared CLI
  def _get_cloudflared(this):
    url = 'https://github.com/cloudflare/cloudflared/releases/' \
          'latest/download/cloudflared-linux-amd64'
    urllib.request.urlretrieve(url, 'cloudflared')
    ret = os.chmod('./cloudflared', 0o775)




from .env import Environment, Section
from .pr import Project
from .sd import SystemDescription
from .ra import RiskAssessment
from .va import VulnerabilityAssessment
from .ex import Exercises
from .log import MANBLogHandler